/**
 * 
 * Stepping program for backlash
 * 
 * Program for sending signals STEP and DIR
 * to the stepper motor driver according to received arguments
 * without changing of the relative position information
 * 
 * Copyright 2019 Luděk Horáček <12@12in.cz>
 * 
 * License GPL
 *
 */

#include <sys/mman.h>
#include <pthread.h>
#include <bcm2835.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <sched.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifndef SELAXIS
#define SELAXIS "axis1.h"
#endif

#include SELAXIS

signed long int clock_ndiff(struct timespec start, struct timespec end);
int d = 0; // direction
unsigned short int stop = 0; // signal

void sig_handler(int signo) {
 if (signo == SIGTERM) stop = 1;
}

int main(int argc, char **argv) {
 if (!bcm2835_init()) return 1;
 if (signal(SIGTERM, sig_handler) == SIG_ERR) printf("\ncan't catch SIGTERM\n");
 if (signal(SIGUSR1, sig_handler) == SIG_ERR) printf("\ncan't catch SIGUSR1\n");

 int fd;
 char *ptr;
 signed long int y = 0, yy = 0; // diff
 unsigned long int i = 0, c = 0; // count of steps
 unsigned long int s = 3200; // ns sleep taken from speed - pulses per second
 unsigned long int a, amax = 0; // acceleration, position of S curve center

 FILE *fptr;
 struct sched_param sp;
 struct timespec tstart={0,0}, slp={0,100}, pau={0,0}, tact={0,0};
 char cfch[256];

 pid_t pid = getpid();
 sprintf(cfch, "%i", pid);
 fptr = fopen(FREEID, "w");
 fwrite(&cfch, sizeof(cfch), 1, fptr); 
 fclose(fptr);
 
  switch (argc) { // utilize arguments
  case 5: // acceleration arg
   a = strtoul(argv[4], &ptr, 10); // 0 = full speed from start to the end
  case 4: // speed arg
   s = 1000000000 / strtoul(argv[3], &ptr, 10); // count pulse interval from speed in steps per second
   if (s <= 3200) s = 3200; // the shortest pulse interval in ns
  case 3: // direction arg
   d = atoi(argv[2]); // 0 = forward
  case 2: // how many steps to do
   i = strtoul(argv[1], &ptr, 10);
 }
 
 if (0 < a) { //  acceleration argument convert to number of steps
  if (2 * a > i) {
   a = i / 2;
  }
  else if (20 >= a) {
   a = a * i / 100;
  }
  amax = a / 10; // position of maximal acceleration
 }

 // high priority in system
 memset(&sp, 0, sizeof(sp));
 sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
 sched_setscheduler(0, SCHED_FIFO, &sp);
 //mlockall(MCL_CURRENT | MCL_FUTURE);



 // signal DIR
 bcm2835_gpio_fsel(DIRPIN, BCM2835_GPIO_FSEL_OUTP);

 // direction
 switch (d) {
  case 1: // backward
   bcm2835_gpio_write(DIRPIN, DIRBACKWARD);
  break;

  default: // forward
   bcm2835_gpio_write(DIRPIN, DIRFORWARD);
  break;
 }
 
 pau.tv_nsec = 5100; // 5.1 microseconds pause - all drivers listed on next links - 10. Sequence Chart of Control Signals
 nanosleep(&pau, NULL);

 pau.tv_nsec = 1600; // 1.6 microseconds in high state - http://www.cncshop.cz/PDF/M542.pdf or http://www.cncshop.cz/PDF/M752.pdf
 //pau.tv_nsec = 7600; // 7.6  microseconds in high state http://www.cncshop.cz/PDF/DM422C.pdf

 bcm2835_gpio_fsel(STEPPIN, BCM2835_GPIO_FSEL_OUTP);
 clock_gettime(CLOCK_MONOTONIC, &tstart);



 while (i > 0) {
  if (stop) break;

  i--;
  c++;

  // signal STEP
  bcm2835_gpio_write(STEPPIN, HIGH); // high state with constant width
  nanosleep(&pau, NULL);
  bcm2835_gpio_write(STEPPIN, LOW); // acceleration and speed change only gab (low state)

  y = 0;
  yy = 0;

  if (0 < a && (c < a || i < a)) { // acceleration
   // based on function f(x) = x ^ 3
   if (c < amax) {
    yy = 500000 - floor(300000 * pow((double)c / amax, 3));
   }
   else if (c < a) {
    yy = floor(200000 * pow((a - (double)c) / a, 3));
   }
   else if (i < amax) {
    yy = 500000 - floor(300000 * pow((double)i / amax, 3));
   }
   else if (i < a) {
    yy = floor(200000 * pow(((a - (double)i) / a), 3));
   }
  }

  if (0 < yy) {
   slp.tv_nsec = yy;
   nanosleep(&slp, NULL);
   slp.tv_nsec = 100;
  }

  while (y < (yy + s)) {
   nanosleep(&slp, NULL);
   clock_gettime(CLOCK_MONOTONIC, &tact);
   y = clock_ndiff(tstart, tact);
  }

  tstart.tv_nsec += s + yy;
 }

 bcm2835_close();
 unlink(FREEID);
 return 0;
}

signed long int clock_ndiff(struct timespec start, struct timespec end) {
 signed long int d;
 d = end.tv_sec - start.tv_sec;
 if (0 < d) d *= 1000000000;
 d += end.tv_nsec - start.tv_nsec;
 return d;
}
